/*************************************************************/
/* File        : wmsetirc-file.c                             */
/* Author      : Paul Lemmens (paul.lemmens@student.kun.nl)  */
/*             : Dan Scott (dan.scott@acm.org)               */
/* Description : Implements wmsetirc-file.h                  */
/* License     : GNU GPL                                     */
/*************************************************************/

/*
wmseti is a Windowmaker dockapp that displays seti@home client statistics.
Copyright (C) 2001 Paul Lemmens, Dan Scott

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <signal.h>
#include <errno.h>

#include "wmsetirc-file.h"
#include "global.h"
#include "key-val.h"
#include "files-dirs.h"

/***********************************************************************************/
/* Implements initialization of a wmsetirc_t struct. If a default value is
 * not possible, set it to zero or NULL. */
/***********************************************************************************/
wmsetirc_t *init_wmsetirc(wmsetirc_t *rc)
{
	int i = 0;

	/* Default setidir is $HOME/setiathome. Initialize only one setidir;
	 * set remaining to NULL. */
	if ( (strcpy(rc->setidirs[0], (char *)"$HOME/setiathome")) == NULL )
	{
		perror("rc->setidirs[0] in init_wmsetirc()");
		return NULL;
	}
	for ( i = 1; i < NUM_SETIDIRS; i++ )
		strcpy(rc->setidirs[i], (char *)"Empty");
	rc->numdirs = 0;


	/* Default current_dir is set to default setidir[0]. */
	if ( (strcpy(rc->currentdir, rc->setidirs[0])) == NULL )
	{
		perror("rc->currentdir in init_wmsetirc()");
		return NULL;
	}
	rc->currentdirnum = 0;

	/* Default rcfile is ~/.wmsetirc. */
	if ( (strcpy(rc->rcfile, (char *)"$HOME/.wmsetirc")) == NULL )
	{
		perror("rc->rcfile in init_wmsetirc()");
		return NULL;
	}


	/* Default sequence to display is: title, percentage done,
	 * time done, number of units received and sent. */
	if ( (strcpy(rc->sequence[0], (char *)"title")) == NULL )
	{
		perror("rc->sequence[0] in init_wmsetirc()");
		return NULL;
	}
	if ( (strcpy(rc->sequence[1], (char *)"prog")) == NULL )
	{
		perror("rc->sequence[1] in init_wmsetirc()");
		return NULL;
	}
	if ( (strcpy(rc->sequence[2], (char *)"cpu")) == NULL )
	{
		perror("rc->sequence[2] in init_wmsetirc()");
		return NULL;
	}
	if ( (strcpy(rc->sequence[3], (char *)"nwus")) == NULL )
	{
		perror("rc->sequence[3] in init_wmsetirc()");
		return NULL;
	}
	if ( (strcpy(rc->sequence[4], (char *)"nresults")) == NULL )
	{
		perror("rc->sequence[4] in init_wmsetirc()");
		return NULL;
	}

	/* Set default pid to zero; don't flash the pid and set the default
	 * signal to SIGSTOP. */
	rc->pid = 0;
	rc->flashpid = 0;
	rc->signal = SIGSTOP;

	return rc;
}

/***********************************************************************************/
/* find_wmsetirc() looks on several default locations to find a wmsetirc file.
 * If found, it will (if necessary) expand this location to an absolute path,
 * put the result in *file and return a pointer to file. wmsetirc can be found
 * in /etc/wmsetirc (old location), /usr/local/etc/wmsetirc (sample wmsetirc)
 * or $HOME/.wmsetirc. */
/***********************************************************************************/
char *find_wmsetirc(char *file)
{
	char buffer[128];
	FILE *fp;

	/* First try $HOME/.wmsetirc, then /usr/local/etc/wmsetirc and finally
	 * as backup /etc/wmsetirc. */
	if ( (strcpy(buffer, (char *)getenv("HOME"))) == NULL )
	{
		perror("finding $HOME in find_wmsetirc()");
		return NULL;
	}
	strcat(buffer, "/.wmsetirc");
	if ( (fp = fopen(buffer, "r")) == NULL )
	{
		/* No $HOME/.wmsetirc. */
		if ( (strcpy(buffer, "/usr/local/etc/wmsetirc")) == NULL )
		{
			perror("copying /usr/local/etc/wmsetirc to buffer in find_wmsetirc()");
			return NULL;
		}
		if ( (fp = fopen(buffer, "r")) == NULL )
		{
			/* No default sample wmsetirc file as well. */
			if ( (strcpy(buffer, "/etc/wmsetirc")) == NULL )
			{
				perror("copying /etc/wmsetirc to buffer in find_wmsetirc()");
				return NULL;
			}
			if ( (fp = fopen(buffer, "r")) == NULL )
			{
				/* Ok. No configuration file found. Notify and die. */
				if ( DebugLevel >= 2 )
					fprintf(stderr, "\t\tNo configuration file found in $HOME/.wmsetirc, /usr/local/etc/wmsetirc or /etc/wmsetirc.\n");
				return NULL;
			}
			else
				fclose(fp);
		}
		else
			fclose(fp);
	}
	else
		fclose(fp);

	/* Because all paths are entered absolute, no expansion needed. */
	if ( (strcpy(file, buffer)) == NULL )
	{
		perror("copying buffer to file in find_wmsetirc()");
		return NULL;
	}
	if ( DebugLevel >= 2 )
		fprintf(stderr, "\t\tUsing configuration file %s\n", file);

	return file;
}

/***********************************************************************************/
/* parse_wmsetirc() parses a previously in file located wmsetirc and puts its
 * values in rc. Return 0 upon success or -1 else upon failure. */
/***********************************************************************************/
int parse_wmsetirc(const char *file, wmsetirc_t *rc)
{
	/* The keyval_t type can be conveniently used here, as can the split_line()
	 * function from stats.h. */
	FILE *fp;
	char buffer[512];
	keyval_t kv_pair;
	int tmp;

	if ( (fp = fopen(file, "r")) == NULL )
	{
		sprintf(ErrMsg, "opening %s in parse_wmsetirc()", file);
		perror(ErrMsg);
		return -1;
	}
	if ( DebugLevel >= 2 )
		fprintf(stderr, "\t\tOpened %s for reading configuration\n", file);

	/* Now loop through the wmsetirc in fp and read all lines not starting 
	 * blank or starting with a '#'. */
	while ( !feof(fp) )
	{
		if ( !fgets(buffer, sizeof(buffer), fp) && !feof(fp) )
		{
			sprintf(ErrMsg, "reading from %s in parse_wmsetirc()", file);
			perror(ErrMsg);
			fprintf(stderr, "\tLast read buffer was <%s>\n", buffer);
			fclose(fp);
			return -1;
		}

		/* Now use the split_line() and keyval_t here. */
		if ( DebugLevel >= 3 )
			fprintf(stderr, "\t\t\tTrying line %s\n", buffer);
		if ( buffer[0] != '#' && buffer[0] != '\n' )
		{
			kv_pair = split_line("=", buffer);
			if ( is_keyword(kv_pair.keyword, (char *)WMSETIRC_KEYWORDS) != 0 )
			{
				fprintf(stderr, "invalid keyword (%s) found in %s in parse_wmsetirc()\n", kv_pair.keyword, file);
				fclose(fp);
				return -1;
			}

			/* We have a valid keyword now. For some keywords extra action has
			 * to be taken. These are setidir (expand the directory) and for
			 * some keywords (WMSETIRC_SEQ_KEYWORDS) they have to be inserted into
			 * a sequence. */
			if ( strcmp(kv_pair.keyword, "setidir") == 0 )
			{
				/* Using insert_setidir() implies that the argument to setidir is expanded
				 * and checked for existance. We don't need to that here. */
				if ( insert_setidir(kv_pair.value, rc) != 0 )
				{
					fprintf(stderr, "inserting %s in setidirs[] in parse_wmsetirc()\n", kv_pair.value);
					fclose(fp);
					return -1;
				}
			}
			else if ( strcmp(kv_pair.keyword, "flashpid") == 0 )
			{
				/* If the value is anything else than 0, make
				 * it a true. Defaults to no. */
				tmp = atoi(kv_pair.value);
				if ( tmp != 0 )
					rc->flashpid = tmp;
			}
			else 
			{
				/* This must be something for the display in the dockapp. */
				tmp = atoi(kv_pair.value);
				if ( DebugLevel >= 3 )
					fprintf(stderr, "\t\t\tConverted %s to %d\n", kv_pair.value, tmp);
				if ( tmp >= ICON_LINES )
				{
					sprintf(ErrMsg, "%s: invalid sequence value (%s) for keyword %s in parse_wmsetirc()\n", file, kv_pair.value, kv_pair.keyword);
					fclose(fp);
					return -1;
				}
				if ( tmp >= 0 )
				{
					if ( insert_stat(kv_pair.keyword, tmp, rc) != 0 )
					{
						fprintf(stderr, "inserting %s in sequence[%d] in parse_wmsetirc()\n", kv_pair.keyword, tmp);
						fclose(fp);
						return -1;
					}
				}
				else
				{
					fprintf(stderr, "%s: negative value for keyword %s in parse_wmsetirc()\n", file, kv_pair.keyword);
					fclose(fp);
					return -1;
				}
			}
		}

	}
	fclose(fp);

	/* Because the rcfile in wmsetirc_t rc is probably still relative,
	 * we can change this because parameter file is absolute. */
	/* Actually this quite a nasty side effect of the parsing! */
	if ( (strcpy(rc->rcfile, file)) == NULL )
	{
		perror("strcpy(rc->rcfile, file) in parse_wmsetirc()");
		return -1;
	}

	/* Same goes for currentdir. But no side effect this time. */
	if ( strcpy(rc->currentdir, rc->setidirs[0]) == NULL )
	{
		perror("copying setidirs[0] to currentdir in parse_wmsetirc()");
		return -1;
	}

	return 0;
}

/***********************************************************************************/
/* print_wmsetirc is used to print out configuration info if the Verbose
 * flag is present on the commandline. Configuration is extracted from rc. */
/***********************************************************************************/
void print_wmsetirc(wmsetirc_t *rc)
{
	int i = 0;

	printf("wmseti %s is using the following values:\n", VERSION);
	printf("Configuration file:\t\t\t%s\n", rc->rcfile);

	printf("%d SETI@home client directorie(s):\t%s\n", rc->numdirs, rc->setidirs[0]);
	for ( i = 1; i < rc->numdirs; i++ )
		printf("\t\t\t\t\t%s\n", rc->setidirs[i]);

	printf("Current monitored directory:\t\t%s\n", rc->currentdir);
	printf("Current SETI@home client PID:\t\t%d\n", rc->pid);

	if ( DebugLevel )
		printf("Debugging level:\t\t\t%d\n", DebugLevel);

	printf("Flash the PID?\t\t\t\t");
	if ( rc->flashpid )
		printf("yes\n");
	else
		printf("no\n");

	printf("Display statistics:\t\t\t");
	for ( i = 0; i < ICON_LINES; i++ )
		if ( i < (ICON_LINES - 1) )
			printf("%s, ", rc->sequence[i]);
		else
			printf("%s\n", rc->sequence[i]);

	return;
}

/***********************************************************************************/
/* insert_setidir() inserts a newly found setidir dir (check for multiple occurences)
 * into the first available position in rc->setidirs[]. dir can be absolute and
 * relative. insert_setidir() also checks for the existance of the directory in the
 * filesystem. Returns 0 upon success or -1 upon failure. */
/***********************************************************************************/
int insert_setidir(const char *dir, wmsetirc_t *rc)
{
	char *buffer, *d, command[128], buffer2[50];
	FILE *fp;
	int i = 0;

	/* numdirs indicates the last taken position in setidirs[]. */
	if ( DebugLevel >= 2 )
		fprintf(stderr, "\t\tTrying to insert %s in setidirs[%d]\n", dir, rc->numdirs);
	if ( rc->numdirs >= NUM_SETIDIRS )
	{
		fprintf(stderr, "crossing arrayboundaries with %d in insert_setidir()\n", rc->numdirs);
		return -1;
	}

	/* First copy the directory so we can access it locally. */
	if ( (buffer = (char *)malloc( (strlen(dir)+1) * sizeof(char)) ) == NULL )
	{
		perror("allocating for buffer in insert_setidir()");
		return -1;
	}
	if ( strcpy(buffer, dir) == NULL )
	{
		perror("copying directory to buffer in insert_setidir()");
		free(buffer);
		return -1;
	}
	if ( DebugLevel >= 3 )
		fprintf(stderr, "\t\t\tbuffer: %s\n", buffer);

	/* Now see if the directory can be expanded to an absolute path if it
	 * isn't absolute already. */
	if ( (d = (char *)expand_dir(buffer)) == NULL )
	{
		fprintf(stderr, "\texpand_dir failed on %s in insert_setidir()\n", buffer);
		if ( buffer )
			free(buffer);
		return -1;
	}
	else
		buffer = d;
	if ( DebugLevel >= 3 )
		fprintf(stderr, "\t\t\tdir: %s\texpanded: %s\n", buffer, d);

	/* Check if the directory really exists in the current filesystem. */
	if ( existing_dir(buffer) != 0 )
	{
		fprintf(stderr, "%s does not exist in insert_setidir()\n", buffer);
		free(buffer);
		return -1;
	}

	/* Finally check if the seti client in that directory is a v2.x or v3.x client by
	 * parsing the output of a popen(). The 3rd line of the output contains
	 * a line "Version: ..." */
	sprintf(command, "%s/setiathome -version", buffer);
	if ( (fp = popen(command, "r")) == NULL )
	{
		sprintf(ErrMsg, "running %s in insert_setidir()", command);
		perror(ErrMsg);
		free(buffer);
		return -1;
	}
	for ( i = 0; i < 3; i++ )
	{
		if ( !fgets(buffer2, sizeof(buffer2), fp) )
		{
			perror("reading from version output in insert_setidir()");
			free(buffer);
			return -1;
		}
	}
        /* Ensure that we are dealing with a v2.x or v3.x client. If yes, just
         * keep on going. This could be expanded to do  conditional processing
         * based on the version number. */
	if ( strstr(buffer2, "Version: 2") == NULL && strstr(buffer2, "Version: 3") == NULL)
	{
		sprintf(ErrMsg, "retrieving version number from %s in insert_setidir()", buffer2);
		perror(ErrMsg);
		free(buffer);
		return -1;
	}

	/* The directory converted from dir to buffer is now absolute, exists in
	 * the current filesystem and has a v2.x or v3.x seti client in it.
	 * This directory can now be entered safely. */

	/* If the first setidir (setidirs[0]) isn't an absolute path, this is 
	 * probably the default since all paths this function gets are absolute. This 
	 * default can be overwritten. */
	if ( rc->setidirs[0][0] != '/' )
	{
		/* For the time being assume that the user doesn't put double setidirs
       		* in the configuration. Doesn't hurt if he does though. */
		if ( strcpy(rc->setidirs[0], buffer) == NULL )
		{
			sprintf(ErrMsg, "overwriting default dir with %s into rc->setidirs[%d] in insert_setidir()", buffer, 0);
			perror(ErrMsg);
			free(buffer);
			return -1;
		}
		if ( DebugLevel >= 2 )
			fprintf(stderr, "\t\tReplaced default dir with %s into rc->setidirs[%d]\n", buffer, 0);
	}
	else
	{
		if ( strcpy(rc->setidirs[rc->numdirs], buffer) == NULL )
		{
			sprintf(ErrMsg, "copying %s into rc->setidirs[%d] in insert_setidir()", buffer, rc->numdirs);
			perror(ErrMsg);
			free(buffer);
			return -1;
		}
		if ( DebugLevel >= 2 )
			fprintf(stderr, "\t\tAppended dir %s in rc->setidirs[%d]\n", buffer, rc->numdirs);
	}
	rc->numdirs++;

	free(buffer);
	return 0;
}

/***********************************************************************************/
/* insert_stat insert a statistic to be displayed in to rc->sequence on position pos
 * (which is relative to 0). Does not check for overwriting a previously inserted
 * stat. statname is assumed to be a valid stat.
 * Returns 0 upon success or -1 upon failure. */
/***********************************************************************************/
int insert_stat(const char *statname, const int pos, wmsetirc_t *rc)
{
	if ( DebugLevel >= 2 )
		fprintf(stderr, "\t\tInserting stat %s on position %d\n", statname, pos);

	if ( pos >= 0 && pos < ICON_LINES )
	{
		if ( (strcmp(rc->sequence[pos], "Empty")) == 0 && DebugLevel >= 2 )
			fprintf(stderr, "\tOverwriting %s with %s on position %d\n", rc->sequence[pos], statname, pos);

		strcpy(rc->sequence[pos], statname);
	}
	else
	{
		fprintf(stderr, "invalid position %d for statistic %s in insert_stat()\n", pos, statname);
		return -1;
	}

	return 0;
}

/***********************************************************************************/
/* fetch_setidir() fetches a certain directory from rc to monitor into dir. n is 
 * the directory to fetch. Returns the directory. */
/***********************************************************************************/
char *fetch_setidir(const int n, const wmsetirc_t *rc, char *dir)
{
	if ( DebugLevel >= 2 )
		fprintf(stderr, "\t\tGetting setidirs[%d]\n", n);

	if ( n < 0 || n > rc->numdirs || n > NUM_SETIDIRS-1 )
		return NULL;

	if ( (strcpy(dir, rc->setidirs[n])) == NULL )
	{
		sprintf(ErrMsg, "fetching setidirs[%d] in init_wmsetirc()", n);
		perror(ErrMsg);
		return NULL;
	}

	return dir;
}

/***********************************************************************************/
/* fetch_sequence() fetches a statistic to be displayed on line n in the dockapp.
 * The statistics name is copied into stat and stat is returned. */
/***********************************************************************************/
char *fetch_sequence(const int n, const wmsetirc_t *rc, char *stat)
{
	if ( DebugLevel >= 2 )
		fprintf(stderr, "\t\tGetting statistic for line %d\n",  n);
	
	if ( n < 0 || n > ICON_LINES-1 )
		return NULL;

	if ( (strcpy(stat, rc->sequence[n])) == NULL )
	{
		sprintf(ErrMsg, "fetching sequence[%d] in init_wmsetirc()", n);
		perror(ErrMsg);
		return NULL;
	}
}

/***********************************************************************************/
/* scrollsetidirs() scrolls to the next available setidir in setidirs[]. Ensures
 * proper cyclic scrolling through setidirs. Returns 0 upon success and -1 upon
 * failure. */
/***********************************************************************************/
int scrollsetidirs(wmsetirc_t *rc)
{
	if ( DebugLevel >= 2 )
		fprintf(stderr, "\t\tScrolling from dir %d to %d\n", rc->currentdirnum, rc->currentdirnum+1);

	/* currentdirnum indicates index of current dir in setidirs[]. Increase
	 * for next scroll. */
	rc->currentdirnum++;

	/* Check if we need to go to the first dir and then overwrite currentdir
	 * with setidirs[currentdirnum]. */
	if ( rc->currentdirnum > rc->numdirs )
		rc->currentdirnum = 0;
	if ( strcpy(rc->currentdir, rc->setidirs[rc->currentdirnum]) == NULL )
	{
		sprintf(ErrMsg, "copying setidirs[%d] to currentdir in scrollsetidirs()", rc->currentdirnum);
		perror(ErrMsg);
		return -1;
	}

	return 0;
}
